import re
import unittest
from datetime import datetime

from codicefiscale import codicefiscale
from codicefiscale.metadata import (
    __author__,
    __copyright__,
    __description__,
    __license__,
    __title__,
    __version__,
)


class CodiceFiscaleTestCase(unittest.TestCase):
    def test_encode_lastname(self):
        data = [
            {
                "input": "",
                "result": "XXX",
            },
            {
                "input": "Caccamo",
                "result": "CCC",
            },
            {
                "input": "Fò",
                "result": "FOX",
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            self.assertEqual(codicefiscale.encode_lastname(obj["input"]), obj["result"])

    def test_encode_firstname(self):
        data = [
            {
                "input": "",
                "result": "XXX",
            },
            {
                "input": "Alessandro",
                "result": "LSN",
            },
            {
                "input": "Dario",
                "result": "DRA",
            },
            {
                "input": "Fabio",
                "result": "FBA",
            },
            {
                "input": "Giovanni",
                "result": "GNN",
            },
            {
                "input": "Hu",
                "result": "HUX",
            },
            {
                "input": "Maria",
                "result": "MRA",
            },
            {
                "input": "Michele",
                "result": "MHL",
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            self.assertEqual(
                codicefiscale.encode_firstname(obj["input"]), obj["result"]
            )

    def test_encode_birthdate_formats(self):
        data = [
            {
                "input": datetime(1985, 4, 3),
                "result": "85D03",
            },
            {
                "input": "03 04 1985",
                "result": "85D03",
            },
            {
                "input": "03/04/1985",
                "result": "85D03",
            },
            {
                "input": "03-04-1985",
                "result": "85D03",
            },
            {
                "input": "03.04.1985",
                "result": "85D03",
            },
            {
                "input": "3/4/1985",
                "result": "85D03",
            },
            {
                "input": "3-4-1985",
                "result": "85D03",
            },
            {
                "input": "3.4.1985",
                "result": "85D03",
            },
            {
                "input": "1985 04 03",
                "result": "85D03",
            },
            {
                "input": "1985/04/03",
                "result": "85D03",
            },
            {
                "input": "1985-04-03",
                "result": "85D03",
            },
            {
                "input": "1985.04.03",
                "result": "85D03",
            },
            {
                "input": "1985/4/3",
                "result": "85D03",
            },
            {
                "input": "1985-4-3",
                "result": "85D03",
            },
            {
                "input": "1985.4.3",
                "result": "85D03",
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            self.assertEqual(
                codicefiscale.encode_birthdate(obj["input"], "M"), obj["result"]
            )

    def test_encode_birthdate_invalid_arguments(self):
        with self.assertRaises(ValueError):
            codicefiscale.encode_birthdate(None, "M")

        with self.assertRaises(ValueError):
            codicefiscale.encode_birthdate("03/04/1985", None)

        with self.assertRaises(ValueError):
            codicefiscale.encode_birthdate("03/04/1985", "X")

        with self.assertRaises(ValueError):
            codicefiscale.encode_birthdate("1985/1985/1985", "M")

    def test_encode_birthdate_gender(self):
        data = [
            {
                "input": ["03/04/1985", "M"],
                "result": "85D03",
            },
            {
                "input": ["03/04/1985", "F"],
                "result": "85D43",
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            self.assertEqual(
                codicefiscale.encode_birthdate(*obj["input"]), obj["result"]
            )

    def test_encode_birthplace_italy(self):
        data = [
            {
                "input": "Torino, Italy",
                "result": "L219",
            },
            {
                "input": "Torino (TO), Italy",
                "result": "L219",
            },
            {
                "input": "Torino (TO)",
                "result": "L219",
            },
            {
                "input": "Torino",
                "result": "L219",
            },
            {
                "input": "L219",
                "result": "L219",
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            self.assertEqual(
                codicefiscale.encode_birthplace(obj["input"]), obj["result"]
            )

    def test_encode_birthplace_foreign_country(self):
        data = [
            {
                "input": "Lettonia",
                "result": "Z145",
            },
            {
                "input": "Giappone",
                "result": "Z219",
            },
            {
                "input": "Marocco",
                "result": "Z330",
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            self.assertEqual(
                codicefiscale.encode_birthplace(obj["input"]), obj["result"]
            )

    def test_encode_birthplace_invalid_arguments(self):
        with self.assertRaises(ValueError):
            codicefiscale.encode_birthplace(None)

        with self.assertRaises(ValueError):
            codicefiscale.encode_birthplace("Area 51")

    def test_encode_birthplace_invalid_birthdate(self):
        with self.assertRaises(ValueError):
            codicefiscale.encode_birthplace("Torino", "01/01/1888")

    def test_encode_cin(self):
        data = [
            {
                "input": "CCCFBA85D03L219",
                "result": "P",
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            self.assertEqual(codicefiscale.encode_cin(obj["input"]), obj["result"])

    def test_encode_cin_invalid_arguments(self):
        with self.assertRaises(ValueError):
            codicefiscale.encode_cin(None)

        with self.assertRaises(ValueError):
            codicefiscale.encode_cin("CCCFBA85D03")

    def test_encode_issue_37(self):
        # when encoding a Codice Fiscale for a person born in Vignola (MO)
        # encode uses code L884 (for Vignola (TN) between 1920-1929) instead of L885.
        data = {
            "lastname": "Caccamo",
            "firstname": "Fabio",
            "gender": "M",
            "birthdate": "03/04/1885",
            "birthplace": "Vignola",
        }
        code = codicefiscale.encode(**data)
        data = codicefiscale.decode(code)
        birthplace = data["birthplace"]
        self.assertTrue(birthplace is not None)
        self.assertTrue(isinstance(birthplace, dict))
        # print(birthplace)
        self.assertEqual(birthplace["code"], "L885")
        self.assertEqual(birthplace["name"], "Vignola")
        self.assertEqual(birthplace["province"], "MO")

    def test_encode(self):
        data = [
            {
                "input": {
                    "lastname": "Ait Hadda",
                    "firstname": "Saad",
                    "gender": "M",
                    "birthdate": "08/09/1995",
                    "birthplace": "Marocco",
                },
                "result": "THDSDA95P08Z330H",
            },
            {
                "input": {
                    "lastname": "Belousovs",
                    "firstname": "Olegs",
                    "gender": "M",
                    "birthdate": "22/03/1984",
                    "birthplace": "Lettonia",
                },
                "result": "BLSLGS84C22Z145O",
            },
            {
                "input": {
                    "lastname": "Bruno",
                    "firstname": "Giovanni",
                    "gender": "M",
                    "birthdate": "26/02/1971",
                    "birthplace": "Torino",
                },
                "result": "BRNGNN71B26L219T",
            },
            {
                "input": {
                    "lastname": "Caccamo",
                    "firstname": "Fabio",
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
                "result": "CCCFBA85D03L219P",
            },
            {
                "input": {
                    "lastname": "Gomba",
                    "firstname": "Alessandro",
                    "gender": "M",
                    "birthdate": "05/01/1984",
                    "birthplace": "Pinerolo",
                },
                "result": "GMBLSN84A05G674H",
            },
            {
                "input": {
                    "lastname": "Martini",
                    "firstname": "Maria",
                    "gender": "F",
                    "birthdate": "16/12/1983",
                    "birthplace": "Anagni",
                },
                "result": "MRTMRA83T56A269B",
            },
            {
                "input": {
                    "lastname": "Panella",
                    "firstname": "Michele",
                    "gender": "M",
                    "birthdate": "27/10/1979",
                    "birthplace": "San Severo (FG)",
                },
                "result": "PNLMHL79R27I158P",
            },
            {
                "input": {
                    "lastname": "Quatrini",
                    "firstname": "Dario",
                    "gender": "M",
                    "birthdate": "13/09/1971",
                    "birthplace": "Pavia",
                },
                "result": "QTRDRA71P13G388J",
            },
            {
                "input": {
                    "lastname": "Takakura",
                    "firstname": "Yuuki",
                    "gender": "F",
                    "birthdate": "28/02/1987",
                    "birthplace": "Torino",
                },
                "result": "TKKYKU87B68L219F",
            },
            {
                "input": {
                    "lastname": "Rossi",
                    "firstname": "Mario",
                    "gender": "M",
                    "birthdate": "17/02/1950",
                    "birthplace": "Porretta Terme",
                },
                "result": "RSSMRA50B17A558W",
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            code = codicefiscale.encode(**obj["input"])
            self.assertEqual(code, obj["result"])

    def test_decode(self):
        data = [
            {
                "input": "THDSDA95P08Z330H",
                "result": {
                    "gender": "M",
                    "birthdate": "08/09/1995",
                    "birthplace": "Marocco",
                },
            },
            {
                "input": "BLSLGS84C22Z145O",
                "result": {
                    "gender": "M",
                    "birthdate": "22/03/1984",
                    "birthplace": "Lettonia",
                },
            },
            {
                "input": "BRNGNN71B26L219T",
                "result": {
                    "gender": "M",
                    "birthdate": "26/02/1971",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "CCCFBA85D03L219P",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "GMBLSN84A05G674H",
                "result": {
                    "gender": "M",
                    "birthdate": "05/01/1984",
                    "birthplace": "Pinerolo",
                },
            },
            {
                "input": "MRTMRA83T56A269B",
                "result": {
                    "gender": "F",
                    "birthdate": "16/12/1983",
                    "birthplace": "Anagni",
                },
            },
            {
                "input": "PNLMHL79R27I158P",
                "result": {
                    "gender": "M",
                    "birthdate": "27/10/1979",
                    "birthplace": "San Severo",
                },
            },
            {
                "input": "QTRDRA71P13G388J",
                "result": {
                    "gender": "M",
                    "birthdate": "13/09/1971",
                    "birthplace": "Pavia",
                },
            },
            {
                "input": "TKKYKU87B68L219F",
                "result": {
                    "gender": "F",
                    "birthdate": "28/02/1987",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "RSSMRA68A01H501Y",
                "result": {
                    "gender": "M",
                    "birthdate": "01/01/1968",
                    "birthplace": "Roma",
                },
            },
            {
                "input": "RSSMRA50B17A558W",
                "result": {
                    "gender": "M",
                    "birthdate": "17/02/1950",
                    "birthplace": "Porretta Terme",
                },
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):
            result = obj["result"]
            obj_decoded = codicefiscale.decode(obj["input"])
            #  print(obj_decoded)
            gender = obj_decoded.get("gender")
            self.assertFalse(gender is None)
            self.assertEqual(gender, result["gender"])

            birthdate = obj_decoded.get("birthdate")
            self.assertFalse(birthdate is None)
            self.assertEqual(
                birthdate, datetime.strptime(result["birthdate"], "%d/%m/%Y")
            )

            birthplace = obj_decoded.get("birthplace")
            self.assertFalse(birthplace is None)
            self.assertEqual(birthplace["name"].upper(), result["birthplace"].upper())

    def test_decode_invalid_syntax(self):
        # invalid lastname
        with self.assertRaises(ValueError):
            codicefiscale.decode("CC0FBA85X03L219P")

        # invalid firstname
        with self.assertRaises(ValueError):
            codicefiscale.decode("CCCFB085X03L219P")

        # invalid date-year
        with self.assertRaises(ValueError):
            codicefiscale.decode("CCCFBA8XD03L219S")

        # invalid date-month
        with self.assertRaises(ValueError):
            codicefiscale.decode("CCCFBA85X03L219P")

        # invalid date-day
        with self.assertRaises(ValueError):
            codicefiscale.decode("CCCFBA85D00L219P")

    def test_decode_omocodia(self):
        data = [
            {
                "input": "CCCFBA85D03L219P",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "CCCFBA85D03L21VE",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "CCCFBA85D03L2MVP",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "CCCFBA85D03LNMVE",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "CCCFBA85D0PLNMVA",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "CCCFBA85DLPLNMVL",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "CCCFBA8RDLPLNMVX",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
            {
                "input": "CCCFBAURDLPLNMVU",
                "result": {
                    "gender": "M",
                    "birthdate": "03/04/1985",
                    "birthplace": "Torino",
                },
            },
        ]

        for obj in data:
            # with self.subTest(obj=obj):

            code = obj["input"]
            result = obj["result"]
            obj_decoded = codicefiscale.decode(code)

            gender = obj_decoded.get("gender")
            self.assertFalse(gender is None)
            self.assertEqual(gender, result["gender"])

            birthdate = obj_decoded.get("birthdate")
            self.assertFalse(birthdate is None)
            self.assertEqual(
                birthdate, datetime.strptime(result["birthdate"], "%d/%m/%Y")
            )

            birthplace = obj_decoded.get("birthplace")
            self.assertFalse(birthplace is None)
            self.assertEqual(birthplace["name"].upper(), result["birthplace"].upper())

            omocodes = obj_decoded.get("omocodes", [])
            self.assertEqual(128, len(omocodes))

    def test_decode_omocodes(self):
        code = "CCCFBA85D03L219P"
        decoded = codicefiscale.decode(code)
        expected_omocodes = [
            "CCCFBA85D03L219P",
            "CCCFBA85D03L21VE",
            "CCCFBA85D03L2M9A",
            "CCCFBA85D03LN19E",
            "CCCFBA85D0PL219L",
            "CCCFBA85DL3L219A",
            "CCCFBA8RD03L219B",
            "CCCFBAU5D03L219M",
            "CCCFBA85D03L2MVP",
            "CCCFBA85D03LN1VT",
            "CCCFBA85D0PL21VA",
            "CCCFBA85DL3L21VP",
            "CCCFBA8RD03L21VQ",
            "CCCFBAU5D03L21VB",
            "CCCFBA85D03LNM9P",
            "CCCFBA85D0PL2M9W",
            "CCCFBA85DL3L2M9L",
            "CCCFBA8RD03L2M9M",
            "CCCFBAU5D03L2M9X",
            "CCCFBA85D0PLN19A",
            "CCCFBA85DL3LN19P",
            "CCCFBA8RD03LN19Q",
            "CCCFBAU5D03LN19B",
            "CCCFBA85DLPL219W",
            "CCCFBA8RD0PL219X",
            "CCCFBAU5D0PL219I",
            "CCCFBA8RDL3L219M",
            "CCCFBAU5DL3L219X",
            "CCCFBAURD03L219Y",
            "CCCFBA85D03LNMVE",
            "CCCFBA85D0PL2MVL",
            "CCCFBA85DL3L2MVA",
            "CCCFBA8RD03L2MVB",
            "CCCFBAU5D03L2MVM",
            "CCCFBA85D0PLN1VP",
            "CCCFBA85DL3LN1VE",
            "CCCFBA8RD03LN1VF",
            "CCCFBAU5D03LN1VQ",
            "CCCFBA85DLPL21VL",
            "CCCFBA8RD0PL21VM",
            "CCCFBAU5D0PL21VX",
            "CCCFBA8RDL3L21VB",
            "CCCFBAU5DL3L21VM",
            "CCCFBAURD03L21VN",
            "CCCFBA85D0PLNM9L",
            "CCCFBA85DL3LNM9A",
            "CCCFBA8RD03LNM9B",
            "CCCFBAU5D03LNM9M",
            "CCCFBA85DLPL2M9H",
            "CCCFBA8RD0PL2M9I",
            "CCCFBAU5D0PL2M9T",
            "CCCFBA8RDL3L2M9X",
            "CCCFBAU5DL3L2M9I",
            "CCCFBAURD03L2M9J",
            "CCCFBA85DLPLN19L",
            "CCCFBA8RD0PLN19M",
            "CCCFBAU5D0PLN19X",
            "CCCFBA8RDL3LN19B",
            "CCCFBAU5DL3LN19M",
            "CCCFBAURD03LN19N",
            "CCCFBA8RDLPL219I",
            "CCCFBAU5DLPL219T",
            "CCCFBAURD0PL219U",
            "CCCFBAURDL3L219J",
            "CCCFBA85D0PLNMVA",
            "CCCFBA85DL3LNMVP",
            "CCCFBA8RD03LNMVQ",
            "CCCFBAU5D03LNMVB",
            "CCCFBA85DLPL2MVW",
            "CCCFBA8RD0PL2MVX",
            "CCCFBAU5D0PL2MVI",
            "CCCFBA8RDL3L2MVM",
            "CCCFBAU5DL3L2MVX",
            "CCCFBAURD03L2MVY",
            "CCCFBA85DLPLN1VA",
            "CCCFBA8RD0PLN1VB",
            "CCCFBAU5D0PLN1VM",
            "CCCFBA8RDL3LN1VQ",
            "CCCFBAU5DL3LN1VB",
            "CCCFBAURD03LN1VC",
            "CCCFBA8RDLPL21VX",
            "CCCFBAU5DLPL21VI",
            "CCCFBAURD0PL21VJ",
            "CCCFBAURDL3L21VY",
            "CCCFBA85DLPLNM9W",
            "CCCFBA8RD0PLNM9X",
            "CCCFBAU5D0PLNM9I",
            "CCCFBA8RDL3LNM9M",
            "CCCFBAU5DL3LNM9X",
            "CCCFBAURD03LNM9Y",
            "CCCFBA8RDLPL2M9T",
            "CCCFBAU5DLPL2M9E",
            "CCCFBAURD0PL2M9F",
            "CCCFBAURDL3L2M9U",
            "CCCFBA8RDLPLN19X",
            "CCCFBAU5DLPLN19I",
            "CCCFBAURD0PLN19J",
            "CCCFBAURDL3LN19Y",
            "CCCFBAURDLPL219F",
            "CCCFBA85DLPLNMVL",
            "CCCFBA8RD0PLNMVM",
            "CCCFBAU5D0PLNMVX",
            "CCCFBA8RDL3LNMVB",
            "CCCFBAU5DL3LNMVM",
            "CCCFBAURD03LNMVN",
            "CCCFBA8RDLPL2MVI",
            "CCCFBAU5DLPL2MVT",
            "CCCFBAURD0PL2MVU",
            "CCCFBAURDL3L2MVJ",
            "CCCFBA8RDLPLN1VM",
            "CCCFBAU5DLPLN1VX",
            "CCCFBAURD0PLN1VY",
            "CCCFBAURDL3LN1VN",
            "CCCFBAURDLPL21VU",
            "CCCFBA8RDLPLNM9I",
            "CCCFBAU5DLPLNM9T",
            "CCCFBAURD0PLNM9U",
            "CCCFBAURDL3LNM9J",
            "CCCFBAURDLPL2M9Q",
            "CCCFBAURDLPLN19U",
            "CCCFBA8RDLPLNMVX",
            "CCCFBAU5DLPLNMVI",
            "CCCFBAURD0PLNMVJ",
            "CCCFBAURDL3LNMVY",
            "CCCFBAURDLPL2MVF",
            "CCCFBAURDLPLN1VJ",
            "CCCFBAURDLPLNM9F",
            "CCCFBAURDLPLNMVU",
        ]
        self.assertEqual(len(decoded["omocodes"]), 128)
        self.assertEqual(decoded["omocodes"], expected_omocodes)

    def test_decode_with_invalid_birthplace(self):
        code = "FRTMXM74L15D354A"
        valid = codicefiscale.is_valid(code)
        self.assertFalse(valid)

    def test_is_omocode(self):
        self.assertFalse(codicefiscale.is_omocode("CCCFBA85D03L219P"))
        self.assertTrue(codicefiscale.is_omocode("CCCFBA85D03L21VE"))
        self.assertTrue(codicefiscale.is_omocode("CCCFBA85D03L2MVP"))
        self.assertTrue(codicefiscale.is_omocode("CCCFBA85D03LNMVE"))
        self.assertTrue(codicefiscale.is_omocode("CCCFBA85D0PLNMVA"))
        self.assertTrue(codicefiscale.is_omocode("CCCFBA85DLPLNMVL"))
        self.assertTrue(codicefiscale.is_omocode("CCCFBA8RDLPLNMVX"))
        self.assertTrue(codicefiscale.is_omocode("CCCFBAURDLPLNMVU"))

    def test_is_valid(self):
        self.assertTrue(codicefiscale.is_valid("CCCFBA85D03L219P"))
        self.assertTrue(codicefiscale.is_valid("CCC FBA 85 D03 L219 P"))
        self.assertTrue(codicefiscale.is_valid("CCC-FBA-85-D03-L219-P"))

        self.assertFalse(codicefiscale.is_valid("CCCFBA85D03L219PP"))  # too long
        self.assertFalse(codicefiscale.is_valid("CCCFBA85D03L219B"))  # wrong CIN
        self.assertFalse(codicefiscale.is_valid("CCCFBA85D03L219"))  # too short
        self.assertFalse(
            codicefiscale.is_valid("CCCFBA85D00L219")
        )  # wrong birthdate day
        self.assertFalse(
            codicefiscale.is_valid("CCCFBA85D99L219")
        )  # wrong birthdate day

    def test_issue_16(self):
        """
        Decode return GIRGENTI (soppresso) instead of AGRIGENTO
        """
        data = codicefiscale.decode("LNNFNC80A01A089K")
        expected_birthplace = {
            "code": "A089",
            "province": "AG",
            "name": "AGRIGENTO",
        }
        birthplace = data["birthplace"]
        self.assertEqual(birthplace["code"], expected_birthplace["code"])
        self.assertEqual(birthplace["name"].upper(), expected_birthplace["name"])
        self.assertEqual(birthplace["province"], expected_birthplace["province"])

    def test_metadata(self):
        self.assertTrue(
            all(
                [
                    __author__,
                    __copyright__,
                    __description__,
                    __license__,
                    __title__,
                    __version__,
                ]
            )
        )

    def test_version(self):
        version_pattern = re.compile(r"^(([\d]+)\.([\d]+)\.([\d]+))$")
        self.assertTrue(version_pattern.match(__version__))



__author__ = "Fabio Caccamo"
__copyright__ = "Copyright (c) 2019-present - Fabio Caccamo"
__description__ = (
    "encode / decode italian fiscal codes "
    "- codifica / decodifica del Codice Fiscale italiano."
)
__email__ = "fabio.caccamo@gmail.com"
__license__ = "MIT"
__title__ = "codicefiscale"
__version__ = "0.7.0"



import argparse
import json
import sys
from datetime import datetime
from typing import Any

from codicefiscale import __description__, __version__, codicefiscale


def _encode_from_args(args: argparse.Namespace) -> None:
    try:
        cf = codicefiscale.encode(
            lastname=args.lastname,
            firstname=args.firstname,
            gender=args.gender,
            birthdate=args.birthdate,
            birthplace=args.birthplace,
        )
    except Exception as error:
        sys.stderr.write(f"{error}\n")
    else:
        sys.stdout.write(f"{cf}\n")


def _decode_from_args(args: argparse.Namespace) -> None:
    try:
        cf_data = codicefiscale.decode(args.code)
    except Exception as error:
        sys.stderr.write(f"{error}\n")
    else:
        if not args.omocodes:
            cf_data.pop("omocodes", None)

        def default_encoder(obj: Any) -> Any:
            if isinstance(obj, datetime):
                return obj.isoformat()

        cf_output = json.dumps(
            cf_data,
            default=default_encoder,
            indent=4,
        )
        sys.stdout.write(f"{cf_output}\n")


def run() -> None:
    parser = argparse.ArgumentParser(
        description=__description__,
    )
    parser.add_argument(
        "--version",
        required=False,
        action="store_true",
        help="Show library version",
    )

    subparsers = parser.add_subparsers(
        title="subcommands",
        dest="subcommand",
        description="Choose a command",
        required=False,
    )

    decode_parser = subparsers.add_parser(
        "decode",
        help=(
            "Decode an italian Codice Fiscale. "
            "For more info run: 'python -m codicefiscale decode --help'"
        ),
    )
    decode_parser.add_argument(
        "code",
        help="Codice Fiscale to decode",
    )
    decode_parser.add_argument(
        "--omocodes",
        required=False,
        action="store_true",
        help="Include omocodes list",
    )

    encode_parser = subparsers.add_parser(
        "encode",
        help=(
            "Encode an italian Codice Fiscale. "
            "For more info run: 'python -m codicefiscale encode --help'"
        ),
    )
    encode_parser.add_argument(
        "--firstname",
        required=True,
        help="First name",
    )
    encode_parser.add_argument(
        "--lastname",
        required=True,
        help="Last name",
    )
    encode_parser.add_argument(
        "--gender",
        choices=["m", "M", "f", "F"],
        required=True,
        help="Gender (M/F)",
    )
    encode_parser.add_argument(
        "--birthdate",
        required=True,
        help="Date of birth (DD/MM/YYYY)",
    )
    encode_parser.add_argument(
        "--birthplace",
        required=True,
        help="Place of birth (city, province)",
    )

    args = parser.parse_args()
    run_with_args(args)


def run_with_args(args: argparse.Namespace) -> None:
    if args.subcommand is None and args.version:
        sys.stdout.write(f"{__version__}\n")
        return
    elif args.subcommand == "decode":
        _decode_from_args(args)
        return
    elif args.subcommand == "encode":
        _encode_from_args(args)
        return
    sys.stdout.write("For more info run: 'python -m codicefiscale --help'\n")



import argparse
import json
import subprocess
import unittest
from io import StringIO
from unittest import mock

from codicefiscale import __version__
from codicefiscale.cli import run, run_with_args


class CodiceFiscaleCLITestCase(unittest.TestCase):
    def test_version(self):
        with mock.patch("sys.stdout", new=StringIO()) as fake_output:
            args = argparse.Namespace(
                subcommand=None,
                version=True,
            )
            run_with_args(args)
            output = fake_output.getvalue().strip()
            self.assertEqual(output, __version__)

    def test_version_from_command_line(self):
        cmd = "python -m codicefiscale --version"
        output = subprocess.check_output(cmd, shell=True).decode("UTF-8").strip()
        self.assertEqual(output, __version__)

    def test_main_without_args(self):
        with mock.patch("sys.stdout", new=StringIO()) as fake_output:
            run()
            output = fake_output.getvalue().strip()
            self.assertEqual(
                output, "For more info run: 'python -m codicefiscale --help'"
            )

    def test_main_without_args_from_command_line(self):
        cmd = "python -m codicefiscale"
        output = subprocess.check_output(cmd, shell=True).decode("UTF-8").strip()
        self.assertEqual(output, "For more info run: 'python -m codicefiscale --help'")

    def test_encode(self):
        with mock.patch("sys.stdout", new=StringIO()) as fake_output:
            args = argparse.Namespace(
                firstname="Mario",
                lastname="Rossi",
                gender="M",
                birthdate="01/01/1990",
                birthplace="Roma,RM",
                subcommand="encode",
            )
            run_with_args(args)
            output = fake_output.getvalue().strip()
            self.assertEqual(output, "RSSMRA90A01H501W")

    def test_encode_from_command_line(self):
        cmd = (
            "python -m codicefiscale encode "
            "--firstname 'Mario' "
            "--lastname 'Rossi' "
            "--gender 'M' "
            "--birthdate '01/01/1990' "
            "--birthplace 'Roma' "
        )
        output = subprocess.check_output(cmd, shell=True).decode("UTF-8").strip()
        self.assertEqual(output, "RSSMRA90A01H501W")

    def test_encode_with_wrong_birthplace(self):
        with mock.patch("sys.stderr", new=StringIO()) as fake_output:
            args = argparse.Namespace(
                firstname="Mario",
                lastname="Rossi",
                gender="M",
                birthdate="01/01/1990",
                birthplace="Romaaa,RM",
                subcommand="encode",
            )
            run_with_args(args)
            output = fake_output.getvalue().strip()
            self.assertEqual(
                output,
                (
                    "[codicefiscale] 'birthplace' / 'birthdate' arguments "
                    "('Romaaa,RM' / '01/01/1990') not mapped to code"
                ),
            )

    def test_decode_without_omocodes(self):
        with mock.patch("sys.stdout", new=StringIO()) as fake_output:
            args = argparse.Namespace(
                code="RSSMRA90A01H501W",
                omocodes=False,
                subcommand="decode",
            )
            run_with_args(args)
            output = fake_output.getvalue().strip()
            self.assertEqual(
                output,
                """
{
    "code": "RSSMRA90A01H501W",
    "gender": "M",
    "birthdate": "1990-01-01T00:00:00",
    "birthplace": {
        "active": false,
        "code": "H501",
        "date_created": "1938-07-27T00:00:00",
        "date_deleted": "1990-03-24T00:00:00",
        "name": "Roma",
        "name_alt": "",
        "name_alt_trans": "",
        "name_slugs": [
            "roma"
        ],
        "name_trans": "Roma",
        "province": "RM"
    },
    "raw": {
        "code": "RSSMRA90A01H501W",
        "lastname": "RSS",
        "firstname": "MRA",
        "birthdate": "90A01",
        "birthdate_year": "90",
        "birthdate_month": "A",
        "birthdate_day": "01",
        "birthplace": "H501",
        "cin": "W"
    }
}
                """.strip(),
            )
            output_data = json.loads(output)
            self.assertEqual(
                output_data,
                {
                    "code": "RSSMRA90A01H501W",
                    "gender": "M",
                    "birthdate": "1990-01-01T00:00:00",
                    "birthplace": {
                        "active": False,
                        "code": "H501",
                        "date_created": "1938-07-27T00:00:00",
                        "date_deleted": "1990-03-24T00:00:00",
                        "name": "Roma",
                        "name_alt": "",
                        "name_alt_trans": "",
                        "name_slugs": ["roma"],
                        "name_trans": "Roma",
                        "province": "RM",
                    },
                    "raw": {
                        "code": "RSSMRA90A01H501W",
                        "lastname": "RSS",
                        "firstname": "MRA",
                        "birthdate": "90A01",
                        "birthdate_year": "90",
                        "birthdate_month": "A",
                        "birthdate_day": "01",
                        "birthplace": "H501",
                        "cin": "W",
                    },
                },
            )

    def test_decode_without_omocodes_from_command_line(self):
        cmd = "python -m codicefiscale decode 'RSSMRA90A01H501W'"
        output = subprocess.check_output(cmd, shell=True).decode("UTF-8").strip()

        self.assertEqual(
            output,
            """
{
    "code": "RSSMRA90A01H501W",
    "gender": "M",
    "birthdate": "1990-01-01T00:00:00",
    "birthplace": {
        "active": false,
        "code": "H501",
        "date_created": "1938-07-27T00:00:00",
        "date_deleted": "1990-03-24T00:00:00",
        "name": "Roma",
        "name_alt": "",
        "name_alt_trans": "",
        "name_slugs": [
            "roma"
        ],
        "name_trans": "Roma",
        "province": "RM"
    },
    "raw": {
        "code": "RSSMRA90A01H501W",
        "lastname": "RSS",
        "firstname": "MRA",
        "birthdate": "90A01",
        "birthdate_year": "90",
        "birthdate_month": "A",
        "birthdate_day": "01",
        "birthplace": "H501",
        "cin": "W"
    }
}
            """.strip(),
        )

        output_data = json.loads(output)
        self.assertEqual(
            output_data,
            {
                "code": "RSSMRA90A01H501W",
                "gender": "M",
                "birthdate": "1990-01-01T00:00:00",
                "birthplace": {
                    "active": False,
                    "code": "H501",
                    "date_created": "1938-07-27T00:00:00",
                    "date_deleted": "1990-03-24T00:00:00",
                    "name": "Roma",
                    "name_alt": "",
                    "name_alt_trans": "",
                    "name_slugs": ["roma"],
                    "name_trans": "Roma",
                    "province": "RM",
                },
                "raw": {
                    "code": "RSSMRA90A01H501W",
                    "lastname": "RSS",
                    "firstname": "MRA",
                    "birthdate": "90A01",
                    "birthdate_year": "90",
                    "birthdate_month": "A",
                    "birthdate_day": "01",
                    "birthplace": "H501",
                    "cin": "W",
                },
            },
        )

    def test_decode_with_omocodes(self):
        with mock.patch("sys.stdout", new=StringIO()) as fake_output:
            args = argparse.Namespace(
                code="RSSMRA90A01H501W",
                omocodes=True,
                subcommand="decode",
            )
            run_with_args(args)
            output = fake_output.getvalue().strip()
            self.assertEqual(
                output,
                """
{
    "code": "RSSMRA90A01H501W",
    "omocodes": [
        "RSSMRA90A01H501W",
        "RSSMRA90A01H50MO",
        "RSSMRA90A01H5L1H",
        "RSSMRA90A01HR01R",
        "RSSMRA90A0MH501O",
        "RSSMRA90AL1H501H",
        "RSSMRA9LA01H501H",
        "RSSMRAV0A01H501L",
        "RSSMRA90A01H5LMZ",
        "RSSMRA90A01HR0MJ",
        "RSSMRA90A0MH50MG",
        "RSSMRA90AL1H50MZ",
        "RSSMRA9LA01H50MZ",
        "RSSMRAV0A01H50MD",
        "RSSMRA90A01HRL1C",
        "RSSMRA90A0MH5L1Z",
        "RSSMRA90AL1H5L1S",
        "RSSMRA9LA01H5L1S",
        "RSSMRAV0A01H5L1W",
        "RSSMRA90A0MHR01J",
        "RSSMRA90AL1HR01C",
        "RSSMRA9LA01HR01C",
        "RSSMRAV0A01HR01G",
        "RSSMRA90ALMH501Z",
        "RSSMRA9LA0MH501Z",
        "RSSMRAV0A0MH501D",
        "RSSMRA9LAL1H501S",
        "RSSMRAV0AL1H501W",
        "RSSMRAVLA01H501W",
        "RSSMRA90A01HRLMU",
        "RSSMRA90A0MH5LMR",
        "RSSMRA90AL1H5LMK",
        "RSSMRA9LA01H5LMK",
        "RSSMRAV0A01H5LMO",
        "RSSMRA90A0MHR0MB",
        "RSSMRA90AL1HR0MU",
        "RSSMRA9LA01HR0MU",
        "RSSMRAV0A01HR0MY",
        "RSSMRA90ALMH50MR",
        "RSSMRA9LA0MH50MR",
        "RSSMRAV0A0MH50MV",
        "RSSMRA9LAL1H50MK",
        "RSSMRAV0AL1H50MO",
        "RSSMRAVLA01H50MO",
        "RSSMRA90A0MHRL1U",
        "RSSMRA90AL1HRL1N",
        "RSSMRA9LA01HRL1N",
        "RSSMRAV0A01HRL1R",
        "RSSMRA90ALMH5L1K",
        "RSSMRA9LA0MH5L1K",
        "RSSMRAV0A0MH5L1O",
        "RSSMRA9LAL1H5L1D",
        "RSSMRAV0AL1H5L1H",
        "RSSMRAVLA01H5L1H",
        "RSSMRA90ALMHR01U",
        "RSSMRA9LA0MHR01U",
        "RSSMRAV0A0MHR01Y",
        "RSSMRA9LAL1HR01N",
        "RSSMRAV0AL1HR01R",
        "RSSMRAVLA01HR01R",
        "RSSMRA9LALMH501K",
        "RSSMRAV0ALMH501O",
        "RSSMRAVLA0MH501O",
        "RSSMRAVLAL1H501H",
        "RSSMRA90A0MHRLMM",
        "RSSMRA90AL1HRLMF",
        "RSSMRA9LA01HRLMF",
        "RSSMRAV0A01HRLMJ",
        "RSSMRA90ALMH5LMC",
        "RSSMRA9LA0MH5LMC",
        "RSSMRAV0A0MH5LMG",
        "RSSMRA9LAL1H5LMV",
        "RSSMRAV0AL1H5LMZ",
        "RSSMRAVLA01H5LMZ",
        "RSSMRA90ALMHR0MM",
        "RSSMRA9LA0MHR0MM",
        "RSSMRAV0A0MHR0MQ",
        "RSSMRA9LAL1HR0MF",
        "RSSMRAV0AL1HR0MJ",
        "RSSMRAVLA01HR0MJ",
        "RSSMRA9LALMH50MC",
        "RSSMRAV0ALMH50MG",
        "RSSMRAVLA0MH50MG",
        "RSSMRAVLAL1H50MZ",
        "RSSMRA90ALMHRL1F",
        "RSSMRA9LA0MHRL1F",
        "RSSMRAV0A0MHRL1J",
        "RSSMRA9LAL1HRL1Y",
        "RSSMRAV0AL1HRL1C",
        "RSSMRAVLA01HRL1C",
        "RSSMRA9LALMH5L1V",
        "RSSMRAV0ALMH5L1Z",
        "RSSMRAVLA0MH5L1Z",
        "RSSMRAVLAL1H5L1S",
        "RSSMRA9LALMHR01F",
        "RSSMRAV0ALMHR01J",
        "RSSMRAVLA0MHR01J",
        "RSSMRAVLAL1HR01C",
        "RSSMRAVLALMH501Z",
        "RSSMRA90ALMHRLMX",
        "RSSMRA9LA0MHRLMX",
        "RSSMRAV0A0MHRLMB",
        "RSSMRA9LAL1HRLMQ",
        "RSSMRAV0AL1HRLMU",
        "RSSMRAVLA01HRLMU",
        "RSSMRA9LALMH5LMN",
        "RSSMRAV0ALMH5LMR",
        "RSSMRAVLA0MH5LMR",
        "RSSMRAVLAL1H5LMK",
        "RSSMRA9LALMHR0MX",
        "RSSMRAV0ALMHR0MB",
        "RSSMRAVLA0MHR0MB",
        "RSSMRAVLAL1HR0MU",
        "RSSMRAVLALMH50MR",
        "RSSMRA9LALMHRL1Q",
        "RSSMRAV0ALMHRL1U",
        "RSSMRAVLA0MHRL1U",
        "RSSMRAVLAL1HRL1N",
        "RSSMRAVLALMH5L1K",
        "RSSMRAVLALMHR01U",
        "RSSMRA9LALMHRLMI",
        "RSSMRAV0ALMHRLMM",
        "RSSMRAVLA0MHRLMM",
        "RSSMRAVLAL1HRLMF",
        "RSSMRAVLALMH5LMC",
        "RSSMRAVLALMHR0MM",
        "RSSMRAVLALMHRL1F",
        "RSSMRAVLALMHRLMX"
    ],
    "gender": "M",
    "birthdate": "1990-01-01T00:00:00",
    "birthplace": {
        "active": false,
        "code": "H501",
        "date_created": "1938-07-27T00:00:00",
        "date_deleted": "1990-03-24T00:00:00",
        "name": "Roma",
        "name_alt": "",
        "name_alt_trans": "",
        "name_slugs": [
            "roma"
        ],
        "name_trans": "Roma",
        "province": "RM"
    },
    "raw": {
        "code": "RSSMRA90A01H501W",
        "lastname": "RSS",
        "firstname": "MRA",
        "birthdate": "90A01",
        "birthdate_year": "90",
        "birthdate_month": "A",
        "birthdate_day": "01",
        "birthplace": "H501",
        "cin": "W"
    }
}
                """.strip(),
            )
            output_data = json.loads(output)
            self.assertEqual(
                output_data,
                {
                    "code": "RSSMRA90A01H501W",
                    "omocodes": [
                        "RSSMRA90A01H501W",
                        "RSSMRA90A01H50MO",
                        "RSSMRA90A01H5L1H",
                        "RSSMRA90A01HR01R",
                        "RSSMRA90A0MH501O",
                        "RSSMRA90AL1H501H",
                        "RSSMRA9LA01H501H",
                        "RSSMRAV0A01H501L",
                        "RSSMRA90A01H5LMZ",
                        "RSSMRA90A01HR0MJ",
                        "RSSMRA90A0MH50MG",
                        "RSSMRA90AL1H50MZ",
                        "RSSMRA9LA01H50MZ",
                        "RSSMRAV0A01H50MD",
                        "RSSMRA90A01HRL1C",
                        "RSSMRA90A0MH5L1Z",
                        "RSSMRA90AL1H5L1S",
                        "RSSMRA9LA01H5L1S",
                        "RSSMRAV0A01H5L1W",
                        "RSSMRA90A0MHR01J",
                        "RSSMRA90AL1HR01C",
                        "RSSMRA9LA01HR01C",
                        "RSSMRAV0A01HR01G",
                        "RSSMRA90ALMH501Z",
                        "RSSMRA9LA0MH501Z",
                        "RSSMRAV0A0MH501D",
                        "RSSMRA9LAL1H501S",
                        "RSSMRAV0AL1H501W",
                        "RSSMRAVLA01H501W",
                        "RSSMRA90A01HRLMU",
                        "RSSMRA90A0MH5LMR",
                        "RSSMRA90AL1H5LMK",
                        "RSSMRA9LA01H5LMK",
                        "RSSMRAV0A01H5LMO",
                        "RSSMRA90A0MHR0MB",
                        "RSSMRA90AL1HR0MU",
                        "RSSMRA9LA01HR0MU",
                        "RSSMRAV0A01HR0MY",
                        "RSSMRA90ALMH50MR",
                        "RSSMRA9LA0MH50MR",
                        "RSSMRAV0A0MH50MV",
                        "RSSMRA9LAL1H50MK",
                        "RSSMRAV0AL1H50MO",
                        "RSSMRAVLA01H50MO",
                        "RSSMRA90A0MHRL1U",
                        "RSSMRA90AL1HRL1N",
                        "RSSMRA9LA01HRL1N",
                        "RSSMRAV0A01HRL1R",
                        "RSSMRA90ALMH5L1K",
                        "RSSMRA9LA0MH5L1K",
                        "RSSMRAV0A0MH5L1O",
                        "RSSMRA9LAL1H5L1D",
                        "RSSMRAV0AL1H5L1H",
                        "RSSMRAVLA01H5L1H",
                        "RSSMRA90ALMHR01U",
                        "RSSMRA9LA0MHR01U",
                        "RSSMRAV0A0MHR01Y",
                        "RSSMRA9LAL1HR01N",
                        "RSSMRAV0AL1HR01R",
                        "RSSMRAVLA01HR01R",
                        "RSSMRA9LALMH501K",
                        "RSSMRAV0ALMH501O",
                        "RSSMRAVLA0MH501O",
                        "RSSMRAVLAL1H501H",
                        "RSSMRA90A0MHRLMM",
                        "RSSMRA90AL1HRLMF",
                        "RSSMRA9LA01HRLMF",
                        "RSSMRAV0A01HRLMJ",
                        "RSSMRA90ALMH5LMC",
                        "RSSMRA9LA0MH5LMC",
                        "RSSMRAV0A0MH5LMG",
                        "RSSMRA9LAL1H5LMV",
                        "RSSMRAV0AL1H5LMZ",
                        "RSSMRAVLA01H5LMZ",
                        "RSSMRA90ALMHR0MM",
                        "RSSMRA9LA0MHR0MM",
                        "RSSMRAV0A0MHR0MQ",
                        "RSSMRA9LAL1HR0MF",
                        "RSSMRAV0AL1HR0MJ",
                        "RSSMRAVLA01HR0MJ",
                        "RSSMRA9LALMH50MC",
                        "RSSMRAV0ALMH50MG",
                        "RSSMRAVLA0MH50MG",
                        "RSSMRAVLAL1H50MZ",
                        "RSSMRA90ALMHRL1F",
                        "RSSMRA9LA0MHRL1F",
                        "RSSMRAV0A0MHRL1J",
                        "RSSMRA9LAL1HRL1Y",
                        "RSSMRAV0AL1HRL1C",
                        "RSSMRAVLA01HRL1C",
                        "RSSMRA9LALMH5L1V",
                        "RSSMRAV0ALMH5L1Z",
                        "RSSMRAVLA0MH5L1Z",
                        "RSSMRAVLAL1H5L1S",
                        "RSSMRA9LALMHR01F",
                        "RSSMRAV0ALMHR01J",
                        "RSSMRAVLA0MHR01J",
                        "RSSMRAVLAL1HR01C",
                        "RSSMRAVLALMH501Z",
                        "RSSMRA90ALMHRLMX",
                        "RSSMRA9LA0MHRLMX",
                        "RSSMRAV0A0MHRLMB",
                        "RSSMRA9LAL1HRLMQ",
                        "RSSMRAV0AL1HRLMU",
                        "RSSMRAVLA01HRLMU",
                        "RSSMRA9LALMH5LMN",
                        "RSSMRAV0ALMH5LMR",
                        "RSSMRAVLA0MH5LMR",
                        "RSSMRAVLAL1H5LMK",
                        "RSSMRA9LALMHR0MX",
                        "RSSMRAV0ALMHR0MB",
                        "RSSMRAVLA0MHR0MB",
                        "RSSMRAVLAL1HR0MU",
                        "RSSMRAVLALMH50MR",
                        "RSSMRA9LALMHRL1Q",
                        "RSSMRAV0ALMHRL1U",
                        "RSSMRAVLA0MHRL1U",
                        "RSSMRAVLAL1HRL1N",
                        "RSSMRAVLALMH5L1K",
                        "RSSMRAVLALMHR01U",
                        "RSSMRA9LALMHRLMI",
                        "RSSMRAV0ALMHRLMM",
                        "RSSMRAVLA0MHRLMM",
                        "RSSMRAVLAL1HRLMF",
                        "RSSMRAVLALMH5LMC",
                        "RSSMRAVLALMHR0MM",
                        "RSSMRAVLALMHRL1F",
                        "RSSMRAVLALMHRLMX",
                    ],
                    "gender": "M",
                    "birthdate": "1990-01-01T00:00:00",
                    "birthplace": {
                        "active": False,
                        "code": "H501",
                        "date_created": "1938-07-27T00:00:00",
                        "date_deleted": "1990-03-24T00:00:00",
                        "name": "Roma",
                        "name_alt": "",
                        "name_alt_trans": "",
                        "name_slugs": ["roma"],
                        "name_trans": "Roma",
                        "province": "RM",
                    },
                    "raw": {
                        "code": "RSSMRA90A01H501W",
                        "lastname": "RSS",
                        "firstname": "MRA",
                        "birthdate": "90A01",
                        "birthdate_year": "90",
                        "birthdate_month": "A",
                        "birthdate_day": "01",
                        "birthplace": "H501",
                        "cin": "W",
                    },
                },
            )

    def test_decode_with_wrong_code(self):
        with mock.patch("sys.stderr", new=StringIO()) as fake_output:
            args = argparse.Namespace(
                code="RSSMRA90A01H501X",
                omocodes=False,
                subcommand="decode",
            )
            run_with_args(args)
            output = fake_output.getvalue().strip()
            self.assertEqual(
                output,
                (
                    "[codicefiscale] wrong CIN (Control Internal Number): "
                    "expected 'W', found 'X'"
                ),
            )



from __future__ import annotations

import re
import string
from datetime import datetime
from itertools import combinations
from typing import Any, Literal, Pattern

import fsutil
from dateutil import parser as date_parser
from slugify import slugify

_CONSONANTS: list[str] = list("bcdfghjklmnpqrstvwxyz")
_VOWELS: list[str] = list("aeiou")
_MONTHS: list[str] = list("ABCDEHLMPRST")
_CIN: dict[str, tuple[int, int]] = {
    "0": (0, 1),
    "1": (1, 0),
    "2": (2, 5),
    "3": (3, 7),
    "4": (4, 9),
    "5": (5, 13),
    "6": (6, 15),
    "7": (7, 17),
    "8": (8, 19),
    "9": (9, 21),
    "A": (0, 1),
    "B": (1, 0),
    "C": (2, 5),
    "D": (3, 7),
    "E": (4, 9),
    "F": (5, 13),
    "G": (6, 15),
    "H": (7, 17),
    "I": (8, 19),
    "J": (9, 21),
    "K": (10, 2),
    "L": (11, 4),
    "M": (12, 18),
    "N": (13, 20),
    "O": (14, 11),
    "P": (15, 3),
    "Q": (16, 6),
    "R": (17, 8),
    "S": (18, 12),
    "T": (19, 14),
    "U": (20, 16),
    "V": (21, 10),
    "W": (22, 22),
    "X": (23, 25),
    "Y": (24, 24),
    "Z": (25, 23),
}
_CIN_REMAINDERS: list[str] = list(string.ascii_uppercase)

_OMOCODIA: dict[str, str] = {
    "0": "L",
    "1": "M",
    "2": "N",
    "3": "P",
    "4": "Q",
    "5": "R",
    "6": "S",
    "7": "T",
    "8": "U",
    "9": "V",
}
_OMOCODIA_DIGITS: str = "".join([digit for digit in _OMOCODIA])
_OMOCODIA_LETTERS: str = "".join([_OMOCODIA[digit] for digit in _OMOCODIA])
_OMOCODIA_ENCODE_TRANS: dict[int, int] = "".maketrans(
    _OMOCODIA_DIGITS, _OMOCODIA_LETTERS
)
_OMOCODIA_DECODE_TRANS: dict[int, int] = "".maketrans(
    _OMOCODIA_LETTERS, _OMOCODIA_DIGITS
)
_OMOCODIA_SUBS_INDEXES: list[int] = list(reversed([6, 7, 9, 10, 12, 13, 14]))
_OMOCODIA_SUBS_INDEXES_COMBINATIONS: list[list[int]] = [[]]
for combo_size in range(1, len(_OMOCODIA_SUBS_INDEXES) + 1):
    for combo in combinations(_OMOCODIA_SUBS_INDEXES, combo_size):
        _OMOCODIA_SUBS_INDEXES_COMBINATIONS.append(list(combo))


def _get_data(filename: str) -> Any:
    return fsutil.read_file_json(fsutil.join_path(__file__, f"data/{filename}"))


def _get_indexed_data() -> (
    dict[str, dict[str, list[dict[str, bool | datetime | str | list[str]]]]]
):
    municipalities = _get_data("municipalities.json")
    countries = _get_data("countries.json")
    data: dict[str, dict[str, list[dict[str, bool | datetime | str | list[str]]]]] = {
        "municipalities": {},
        "countries": {},
        "codes": {},
    }

    for municipality in municipalities:
        code = municipality["code"]
        province = municipality["province"].lower()
        names = municipality["name_slugs"]
        for name in names:
            name_and_province = f"{name}-{province}"
            data["municipalities"].setdefault(name, [])
            data["municipalities"].setdefault(name_and_province, [])
            data["municipalities"][name].append(municipality)
            data["municipalities"][name_and_province].append(municipality)
        data["codes"].setdefault(code, [])
        data["codes"][code].append(municipality)

    for country in countries:
        code = country["code"]
        names = country["name_slugs"]
        for name in names:
            data["countries"].setdefault(name, [])
            data["countries"][name].append(country)
        data["codes"].setdefault(code, [])
        data["codes"][code].append(country)

    return data


_DATA: dict[str, dict[str, list[dict[str, Any]]]] = _get_indexed_data()

CODICEFISCALE_RE: Pattern[str] = re.compile(
    r"^"
    r"(?P<lastname>[a-z]{3})"
    r"(?P<firstname>[a-z]{3})"
    r"(?P<birthdate>(?P<birthdate_year>[a-z\d]{2})(?P<birthdate_month>[abcdehlmprst]{1})(?P<birthdate_day>[a-z\d]{2}))"  # noqa: B950, E501
    r"(?P<birthplace>[a-z]{1}[a-z\d]{3})"
    r"(?P<cin>[a-z]{1})$",
    re.IGNORECASE,
)


def _get_consonants(s: str) -> list[str]:
    return [char for char in s if char in _CONSONANTS]


def _get_vowels(s: str) -> list[str]:
    return [char for char in s if char in _VOWELS]


def _get_consonants_and_vowels(
    consonants: list[str],
    vowels: list[str],
) -> str:
    return "".join(list(consonants[:3] + vowels[:3] + (["X"] * 3))[:3]).upper()


def _get_date(
    date: datetime | str | None,
    separator: str = "-",
) -> datetime | None:
    if not date:
        return None
    if isinstance(date, datetime):
        return date
    date_slug = slugify(date)
    date_parts = date_slug.split("-")[:3]
    date_parser_options = (
        {
            "yearfirst": True,
        }
        if len(date_parts[0]) == 4
        else {
            "dayfirst": True,
        }
    )
    try:
        date_obj = date_parser.parse(
            date_slug,
            parserinfo=date_parser.parserinfo(**date_parser_options),
        )
        return date_obj
    except ValueError:
        return None


def _get_birthplace(
    birthplace: str,
    birthdate: datetime | str | None = None,
) -> dict[str, dict[str, bool | datetime | str | list[str]]] | None | None:
    birthplace_slug = slugify(birthplace)
    birthplace_code = birthplace_slug.upper()
    birthplaces_options = _DATA["municipalities"].get(
        birthplace_slug,
        _DATA["countries"].get(
            birthplace_slug,
            _DATA["codes"].get(
                birthplace_code,
            ),
        ),
    )
    if not birthplaces_options:
        return None

    birthdate_date = _get_date(birthdate)
    if not birthdate_date:
        return birthplaces_options[0].copy()

    for birthplace_option in birthplaces_options:
        date_created = _get_date(birthplace_option["date_created"]) or datetime.min
        date_created = date_created.replace(tzinfo=None)
        date_deleted = _get_date(birthplace_option["date_deleted"]) or datetime.max
        date_deleted = date_deleted.replace(tzinfo=None)
        # print(birthdate_date, date_created, date_deleted)
        if birthdate_date >= date_created and birthdate_date <= date_deleted:
            return birthplace_option.copy()

    return None


def _get_omocode(
    code: str,
    subs: list[int],
    trans: dict[int, int],
) -> str:
    code_chars = list(code[0:15])
    for i in subs:
        code_chars[i] = code_chars[i].translate(trans)
    code = "".join(code_chars)
    code_cin = encode_cin(code)
    code += code_cin
    return code


def _get_omocodes(code: str) -> list[str]:
    code_root = _get_omocode(
        code, subs=_OMOCODIA_SUBS_INDEXES, trans=_OMOCODIA_DECODE_TRANS
    )
    codes = [
        _get_omocode(code_root, subs=subs, trans=_OMOCODIA_ENCODE_TRANS)
        for subs in _OMOCODIA_SUBS_INDEXES_COMBINATIONS
    ]
    return codes


def encode_lastname(lastname: str) -> str:
    """
    Encode lastname to the code used in italian fiscal code.

    :param lastname: The lastname
    :type lastname: string

    :returns: The code used in italian fiscal code
    :rtype: string
    """
    lastname_slug = slugify(lastname)
    lastname_consonants = _get_consonants(lastname_slug)
    lastname_vowels = _get_vowels(lastname_slug)
    lastname_code = _get_consonants_and_vowels(lastname_consonants, lastname_vowels)
    return lastname_code


def encode_firstname(firstname: str) -> str:
    """
    Encodes firstname to the code used in italian fiscal code.

    :param firstname: The firstname
    :type firstname: string

    :returns: The code used in italian fiscal code
    :rtype: string
    """
    firstname_slug = slugify(firstname)
    firstname_consonants = _get_consonants(firstname_slug)

    if len(firstname_consonants) > 3:
        del firstname_consonants[1]

    firstname_vowels = _get_vowels(firstname_slug)
    firstname_code = _get_consonants_and_vowels(firstname_consonants, firstname_vowels)
    return firstname_code


def encode_birthdate(
    birthdate: datetime | str | None,
    gender: Literal["m", "M", "f", "F"],
) -> str:
    """
    Encodes birthdate to the code used in italian fiscal code.

    :param birthdate: The birthdate
    :type birthdate: datetime or string
    :param gender: The gender, 'M' or 'F'
    :type gender: string

    :returns: The code used in italian fiscal code
    :rtype: string
    """
    if not birthdate:
        raise ValueError("[codicefiscale] 'birthdate' argument cant be None")
    date = _get_date(birthdate)
    if not date:
        raise ValueError("[codicefiscale] 'date' argument cant be None")

    if not gender:
        raise ValueError("[codicefiscale] 'gender' argument cant be None")
    gender_code = gender.upper()
    if gender_code not in ("M", "F"):
        raise ValueError("[codicefiscale] 'gender' argument must be 'M' or 'F'")

    year_code = str(date.year)[2:]
    month_code = _MONTHS[date.month - 1]
    day_code = str(date.day + (40 if gender_code == "F" else 0)).zfill(2).upper()
    date_code = f"{year_code}{month_code}{day_code}"
    return date_code


def encode_birthplace(
    birthplace: str,
    birthdate: datetime | str | None = None,
) -> str | None:
    """
    Encodes birthplace to the code used in italian fiscal code.

    :param birthplace: The birthplace
    :type birthplace: string

    :returns: The code used in italian fiscal code
    :rtype: string
    """
    if not birthplace:
        raise ValueError("[codicefiscale] 'birthplace' argument cant be None")

    birthplace_without_province = re.split(r",|\(", birthplace)[0]
    birthplace_data = _get_birthplace(
        birthplace,
        birthdate,
    ) or _get_birthplace(
        birthplace_without_province,
        birthdate,
    )

    if not birthplace_data:
        raise ValueError(
            "[codicefiscale] 'birthplace' / 'birthdate' arguments "
            f"({birthplace!r} / {birthdate!r}) not mapped to code"
        )

    birthplace_code = str(birthplace_data["code"])
    return birthplace_code


def encode_cin(code: str) -> str:
    """
    Encodes cin to the code used in italian fiscal code.

    :param code: The code
    :type code: string

    :returns: The code used in italian fiscal code
    :rtype: string
    """
    if not code:
        raise ValueError("[codicefiscale] 'code' argument cant be None")

    code_len = len(code)
    if code_len not in [15, 16]:
        raise ValueError(
            f"[codicefiscale] 'code' length must be 15 or 16, not: {code_len}"
        )

    cin_tot = 0
    for i, char in enumerate(code[0:15]):
        cin_tot += _CIN[char][int(bool((i + 1) % 2))]
    cin_code = _CIN_REMAINDERS[cin_tot % 26]

    # print(cin_code)
    return cin_code


def encode(
    lastname: str,
    firstname: str,
    gender: Literal["m", "M", "f", "F"],
    birthdate: datetime | str | None,
    birthplace: str,
) -> str:
    """
    Encodes the italian fiscal code.

    :param lastname: The lastname
    :type lastname: string
    :param firstname: The firstname
    :type firstname: string
    :param gender: The gender, 'M' or 'F'
    :type gender: string
    :param birthdate: The birthdate
    :type birthdate: datetime or string
    :param birthplace: The birthplace
    :type birthplace: string

    :returns: The italian fiscal code
    :rtype: string
    """

    lastname_code = encode_lastname(lastname)
    firstname_code = encode_firstname(firstname)
    birthdate_code = encode_birthdate(birthdate, gender)
    birthplace_code = encode_birthplace(birthplace, birthdate)
    code = f"{lastname_code}{firstname_code}{birthdate_code}{birthplace_code}"
    cin_code = encode_cin(code)
    code = f"{code}{cin_code}"

    # raise ValueError if code is not valid
    decode(code)
    return code


def decode_raw(code: str) -> dict[str, str]:
    """
    Decodes the raw data associated to the code.

    :param code: The code
    :type code: string

    :returns: The raw data associated to the code.
    :rtype: dict
    """
    code = slugify(code)
    code = code.replace("-", "")
    code = code.upper()

    match = CODICEFISCALE_RE.match(code)
    if not match:
        raise ValueError(f"[codicefiscale] invalid syntax: {code}")

    data = {
        "code": code,
        "lastname": match["lastname"],
        "firstname": match["firstname"],
        "birthdate": match["birthdate"],
        "birthdate_year": match["birthdate_year"],
        "birthdate_month": match["birthdate_month"],
        "birthdate_day": match["birthdate_day"],
        "birthplace": match["birthplace"],
        "cin": match["cin"],
    }

    return data


def decode(code: str) -> dict[str, Any]:
    """
    Decodes the italian fiscal code.

    :param code: The code
    :type code: string

    :returns: The data associated to the code and some additional info.
    :rtype: dict
    """
    raw = decode_raw(code)

    code = raw["code"]

    birthdate_year = int(raw["birthdate_year"].translate(_OMOCODIA_DECODE_TRANS))
    birthdate_month = _MONTHS.index(raw["birthdate_month"]) + 1
    birthdate_day = int(raw["birthdate_day"].translate(_OMOCODIA_DECODE_TRANS))

    if birthdate_day > 40:
        birthdate_day -= 40
        gender = "F"
    else:
        gender = "M"

    current_year = datetime.now().year
    current_year_century_prefix = str(current_year)[0:-2]
    birthdate_year = int(f"{current_year_century_prefix}{birthdate_year}")
    if birthdate_year > current_year:
        birthdate_year -= 100
    birthdate_str = f"{birthdate_year}/{birthdate_month}/{birthdate_day}"
    birthdate = _get_date(birthdate_str, separator="/")
    if not birthdate:
        raise ValueError(f"[codicefiscale] invalid date: {birthdate_str}")

    birthplace_code = raw["birthplace"][0] + raw["birthplace"][1:].translate(
        _OMOCODIA_DECODE_TRANS
    )
    birthplace = _get_birthplace(birthplace_code, birthdate)
    # print(birthplace)
    if not birthplace:
        raise ValueError(f"[codicefiscale] wrong birthplace code: {birthplace_code!r}")

    cin = raw["cin"]
    cin_check = encode_cin(code)
    # print(cin, cin_check)
    if cin != cin_check:
        raise ValueError(
            "[codicefiscale] wrong CIN (Control Internal Number): "
            f"expected {cin_check!r}, found {cin!r}"
        )

    data = {
        "code": code,
        "omocodes": _get_omocodes(code),
        "gender": gender,
        "birthdate": birthdate,
        "birthplace": birthplace,
        "raw": raw,
    }

    # print(data)
    return data


def is_omocode(code: str) -> bool:
    """
    Determines whether the specified code is omocode or not.

    :param code: The code
    :type code: string

    :returns: True if the specified code is omocode, False otherwise.
    :rtype: boolean
    """
    data = decode(code)
    codes = data["omocodes"]
    codes.pop(0)
    return code in codes


def is_valid(code: str) -> bool:
    """
    Determines whether the specified code is valid.

    :param code: The code
    :type code: string

    :returns: True if the specified code is valid, False otherwise.
    :rtype: boolean
    """
    try:
        decode(code)
        return True
    except ValueError:
        return False
